import datetime
import json
from pathlib import Path
from collections.abc import Mapping, Set
import logging
from typing import Iterable

import dateutil.parser

from .settings import ENCODING, LANGUAGE, START_CHARACTER
from .utils import remove_emoji

logger = logging.getLogger(__name__)

class Tweet:

    # Tweet URIs don't care what username you use
    NULL_USERNAME = '_'

    def __init__(self, data: Mapping, language: str = None, username: str = None):
        # This is the raw data format from the Twitter archive file
        self._data = data
        self._language = language or LANGUAGE
        self.username = username or self.NULL_USERNAME
        self._created_at = None

    @staticmethod
    def load(path: Path, encoding=None, start_char: str = None) -> Iterable[Mapping]:
        encoding = encoding or ENCODING
        start_char = start_char or START_CHARACTER

        # Load input file into memory
        # (Ideally we'd stream this, but the unpredictable format prevents this.)
        with path.open(encoding=encoding) as file:
            data = file.read()

        # Remove everything before the first square bracket
        data = data[data.find(start_char):]

        # Parse JSON
        # Each object in the document has a 'tweet' key
        for doc in json.loads(data):
            yield doc['tweet']

    @property
    def language(self) -> str:        

        # Language code ISO 639-1
        lang = self._data.get('lang', self._language)

        # Remove undetermined language (any language match) or artificial language
        if lang in {'und', 'art'}:
            lang = None

        return lang

    @property
    def created_at(self) -> datetime.datetime:
        if self._created_at is None:
            self._created_at = dateutil.parser.parse(self._data['created_at'])
        return self._created_at

    @property
    def urls(self) -> Mapping[str, str]:
        """
        Map Twitter short-URLs to full URLs so they can be replaced in the
        main Tweet body text.
        """
        urls = dict()
        
        try:
            for media in self._data['extended_entities']['media']:
                urls[media['url']] = media['media_url']
        except KeyError:
            pass

        for url in self._data['entities']['urls']:
            urls[url['url']] = url['expanded_url']

        return urls

    @property
    def images(self) -> set:
        images = set()

        # Get media URLs (these also map to short urls) and add these to the short-URL map
        # This media key might not exist
        try:
            for media in self._data['extended_entities']['media']:
                images.add(media['media_url'])
        except KeyError:
            pass

        return images

    @property
    def full_text(self) -> str:
        
        # Process message body
        # Remove emoji characters (these don't convert to PDF very well)
        full_text = remove_emoji(self._data['full_text'], lang=self.language)

        # Remove crazy characters, which will also break PDF
        # https://docs.python.org/3/howto/unicode.html
        try:
            full_text = full_text.encode('utf-8').decode('unicode-escape')
        except UnicodeDecodeError as exc:
            logger.error(exc)
            logger.warning(full_text)
            pass

        # Convert short urls (t.co) into their original links
        for short_url, expanded_url in self.urls.items():
            full_text = full_text.replace(short_url, expanded_url)

        return full_text

    @property
    def _hashtags(self):
        return self._data['entities']['hashtags']

    @property
    def hashtags(self) -> Set[str]:
        return {hashtag['text'].casefold() for hashtag in self._hashtags}

    @property
    def id_str(self) -> str:
        return self._data['id_str']

    @property
    def uri(self) -> str:
        """
        The full web address of this tweet.
        """
        return f"https://twitter.com/{self.username}/status/{self.id_str}"



import logging
import urllib.parse
import tempfile
from pathlib import Path
from contextlib import contextmanager

import requests
import emoji
import requests.adapters

from .settings import LANGUAGE

logger = logging.getLogger(__name__)

@contextmanager
@staticmethod
def download(uri: str, session: requests.Session):

    logger.info(f'Downloading "{uri}"...')
    with session.get(uri) as response:
        response.raise_for_status()
        yield response
        

@contextmanager
def download_temp_file(uri: str, session: requests.Session):   

    # Get filename from URL
    filename = Path(urllib.parse.urlsplit(uri).path).name

    with download(uri, session=session) as response:
        # Store file on local disk
        with tempfile.TemporaryDirectory() as directory:
            with Path(directory).joinpath(filename).open('wb') as file:
                file.write(response.content)
                yield file


def remove_emoji(s: str, lang: str = None) -> str:
    """
    Remove emojis, which could break PDF output

    https://carpedm20.github.io/emoji/docs/#replacing-and-removing-emoji
    """

    def replace(s, data_dict):
        nonlocal lang

        try:
            return data_dict[lang]
        # If this language isn't there, default
        except KeyError:
            return data_dict[LANGUAGE]

    # Replace emoji symbols with text representation
    return emoji.replace_emoji(s, replace=replace)


def get_session(**kwargs) -> requests.Session:
    session = requests.Session()

    # Configure backoff strategy
    retry = requests.adapters.Retry(total=3, backoff_factor=1)
    adapter = requests.adapters.HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter=adapter)
    session.mount('https://', adapter=adapter)

    return session



#!/usr/bin/env python3

import argparse
import logging
import datetime
import itertools
from typing import Optional

from pathlib import Path

import dateutil.parser
import pytz
import requests

from .tweet import Tweet
from .pdf import PDFDocument
from .settings import ENCODING, LANGUAGE, FONT_FAMILY, FONT_SIZE, FONT, IMAGE_WIDTH
import tweets2pdf.utils

DESCRIPTION = """
Parse Twitter archived JSON files and convert to PDF files
"""


logger = logging.getLogger(__name__)


def utc_timestamp(timestamp: str) -> datetime.datetime:
    """
    Parse user-inputted timestamp
    """
    # Parse timestamp
    t = dateutil.parser.parse(timestamp)
    # Convert to UTC
    return pytz.UTC.localize(t)


def get_args():
    """
    Command-line arguments
    """

    parser = argparse.ArgumentParser(description=DESCRIPTION)

    # Twitter archive options
    parser.add_argument("-f", "--file", type=Path, dest="filename", help="Path to Twitter JSON archive", required=True)
    parser.add_argument('--encoding', default=ENCODING, help='Input file character set')
    parser.add_argument("-a", "--hashtags", help="Filter by hashtags (comma-separated list)", default=str())
    parser.add_argument("-s", "--date_start", dest="date_start", type=utc_timestamp)
    parser.add_argument("-e", "--date_end", dest="date_end", type=utc_timestamp)
    parser.add_argument('--start', type=int, help="Start at this tweet number")
    parser.add_argument('--stop', type=int, help="Stop after this many tweets")

    # PDF output options
    parser.add_argument("-p", "--pdf", type=Path, help='Output PDF file path', required=True)
    parser.add_argument('-i', '--images', help='Download images and put in the PDF document (SLOW)', action='store_true')
    parser.add_argument('-w', '--imgwidth', help="Image width", default=IMAGE_WIDTH)
    
    # Font options
    parser.add_argument('--font', help='Unicode font TTF file', type=Path, default=FONT)
    parser.add_argument('--font_family', default=FONT_FAMILY)
    parser.add_argument('--font_size', type=float, default=FONT_SIZE)
    parser.add_argument('--loglevel', default='INFO', help="Verbosity level: DEBUG, INFO, WARNING, ERROR")
    parser.add_argument('--language', default=LANGUAGE, help='Language ISO 639-1 code')

    return parser.parse_args()


def main():
    options = get_args()
    logging.basicConfig(level=options.loglevel)

    # Get the selected hashtags
    hashtag_filter = {s.strip() for s in options.hashtags.casefold().split(',')}

    pdf = PDFDocument(font=options.font, font_family=options.font_family, font_size=options.font_size)

    # Create a HTTP session if we want to download images
    session: Optional[requests.Session] = None
    if options.images:
        session = tweets2pdf.utils.get_session()

    # Iterate over input tweets
    tweet_count = 0
    for tweet_data in itertools.islice(Tweet.load(options.filename, encoding=options.encoding), options.start, options.stop):
        tweet_count += 1

        # Show progress
        if tweet_count % 1000 == 0:
            logger.info(f"Processing tweet {tweet_count}")

        # Decode tweet in a simple structure
        tweet = Tweet(tweet_data)

        # Time filter
        try:
            if tweet.created_at < options.date_start:
                continue
        except TypeError:
            pass

        try:
            if tweet.created_at >= options.date_end:
                continue
        except TypeError:
            pass

        # Filter by hashtag
        if not tweet.hashtags.union(hashtag_filter):
            continue

        pdf.add_tweet(tweet, download_images=options.images, image_width=options.imgwidth, session=session)

    pdf.output(options.pdf)

    if session is not None:
        session.close()


if __name__ == '__main__':
    main()
